/* NickServ functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean
 *     E-mail: auspice@pleaselook.com
 *
 * SirvNET Services is copyright (c) 1998-2000 Trevor Klingbeil.
 *     E-mail: <priority1@dal.net>
 * Originally based on EsperNet Services(c) by Andy Church.
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "../inc/services.h"
#include "../inc/timeout.h"
#include "../inc/encrypt.h"

static NickInfo *nicklists[256];	/* One for each initial character */

#include "ns-help.c"


/* For local timeout use: */
#define TO_COLLIDE	0	/* Collide the user with this nick */
#define TO_RELEASE	1	/* Release a collided nick */
#define TO_COLLIDE_TL   2       /* Time left warning to collide */

static void alpha_insert_nick(NickInfo *ni);
static NickInfo *makenick(const char *nick);
static int delnick(NickInfo *ni);
static void do_unhold(const char *source, const char *nick);
static void do_unmark(const char *source, const char *nick);

static void collide(NickInfo *ni, int from_timeout);
static void release(NickInfo *ni, int from_timeout);
static void add_ns_timeout(NickInfo *ni, int type, time_t delay);
static void del_ns_timeout(NickInfo *ni, int type);

static void update_last_seen_time(Timeout *to);

static void do_set_malay(NickInfo *ni, char *param);
static void do_set_auth(NickInfo *ni, char *param);
static void do_set_privmsg(NickInfo *ni, char *param);
static void do_ohelp(const char *source);
static void do_help(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_delete(const char *source);
static void do_drop(const char *source);
static void do_set(const char *source);
static void do_set_password(NickInfo *ni, char *param);
static void do_set_url(NickInfo *ni, char *param);
static void do_set_email(NickInfo *ni, char *param);
static void do_set_kill(NickInfo *ni, char *param);
static void do_set_secure(NickInfo *ni, char *param);
static void do_set_private(NickInfo *ni, char *param);
static void do_set_emailmemos(NickInfo *ni, char *param);
static void do_set_neverop(NickInfo *ni, char *param);
static void do_set_noop(NickInfo *ni, char *param);
static void do_set_nomemo(NickInfo *ni, char *param);
static void do_set_nohost(NickInfo *ni, char *param);
static void do_access(const char *source);
// static void do_link(const char *source);
static void do_info(const char *source);
static void do_acc(const char *source);
static void do_flist(const char *source);
static void do_list(const char *source);
static void do_recover(const char *source);
static void do_release(const char *source);
static void do_ghost(const char *source);
static void do_getpass(const char *source);
static void do_sendpass(const char *source);
static void do_forbid(const char *source);
static void do_unforbid(const char *source, const char *nick);
static char *get_os_access(NickInfo *ni);
static void do_setpass(const char *source);
static void do_set_hidemail(NickInfo *ni, char *param);
static void do_setflags(const char *source);

/*************************************************************************/

#ifndef SKELETON

/* Display total number of registered nicks and info about each; or, if
 * a specific nick is given, display information about that nick (like
 * /msg NickServ INFO <nick>).  If count_only != 0, then only display the
 * number of registered nicks (the nick parameter is ignored).
 */

void listnicks(int count_only, const char *nick)
{
    long count = 0;
    NickInfo *ni;
    int i;

    if (count_only) {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next)
		++count;
	}
	printf("%d nicknames registered.\n", count);

    } else if (nick) {

	struct tm tm;
	char buf[512];

	if (!(ni = findnick(nick))) {
	    printf("%s not registered.\n");
	    return;
	} else if (ni->flags & NI_VERBOTEN) {
	    printf("%s is FORBIDden.\n");
	    return;
	}
	printf("%s is %s\n", nick, ni->last_realname);
	printf("Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);

	printf("  Time registered: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year);
	tm = *localtime(&ni->last_seen);
	printf("   Last seen time: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year);
	if (ni->url)
	    printf("              URL: %s\n", ni->url);
        if (ni->email && (ni->flags & NI_HIDE_EMAIL))
            printf("   E-mail address: Hidden\n");
        else
	    printf("   E-mail address: %s\n", ni->email);
	*buf = 0;
	if (ni->flags & NI_KILLPROTECT)
	    strcat(buf, "Kill protection");
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
	if (ni->flags & NI_PRIVATE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Private");
	}
        if (ni->flags & NI_MARK) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Marked");
        }
        if (ni->flags & NI_HOLD) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Held");
        }
	if (!*buf)
	    strcpy(buf, "None");
	printf("          Options: %s\n", buf);

    } else {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		printf("    %-20s  %s\n", ni->nick, ni->last_usermask);
		++count;
	    }
	}
	printf("%d nicknames registered.\n", count);

    }
}

#endif	/* !SKELETON */

/*************************************************************************/

NickInfo *retnick(int i)
{
    return nicklists[i];
}

/* Return information on memory use.  Assumes pointers are valid. */

void get_nickserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    NickInfo *ni;
    char **accptr;


    for (i = 0; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    count++;
	    mem += sizeof(*ni);
	    if (ni->last_usermask)
		mem += strlen(ni->last_usermask)+1;
	    if (ni->last_realname)
		mem += strlen(ni->last_realname)+1;
            if (ni->url)
                mem += strlen(ni->url)+1;
            if (ni->email)
                mem += strlen(ni->email)+1;
            if (ni->forward)
                mem += strlen(ni->forward)+1;
            if (ni->mark)
                mem += strlen(ni->mark)+1;
            if (ni->hold)
                mem += strlen(ni->hold)+1;
            if (ni->forbid)
                mem += strlen(ni->forbid)+1;

	    mem += sizeof(char *) * ni->accesscount;
	    for (accptr=ni->access, j=0; j < ni->accesscount; accptr++, j++) {
		if (*accptr)
		    mem += strlen(*accptr)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/
/*************************************************************************/

/* Main NickServ routine. */

void nickserv(const char *source, char *buf)
{
    char *cmd, *s;
    NickInfo *ni = findnick(source);

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_NickServ, source, "\1PING %s", s);

#ifdef SKELETON

    } else {
	notice(s_NickServ, source, "%s is currently offline.", s_NickServ);
        return;

    }

#else

    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
    
        if (!(s = strtok (NULL, "")))
            s = "\1";
        notice(s_NickServ, source, "%sVERSION Auspice IRC Services %s", s, s);

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "REGISTER") == 0) {
	do_register(source);

    } else if (stricmp(cmd, "IDENTIFY") == 0) {
	do_identify(source);

    } else if (stricmp(cmd, "ID") == 0) {
        do_identify(source);

/*    } else if (stricmp(cmd, "LINK") == 0) {
	do_link(source);
*/

    } else if (stricmp(cmd, "DROP") == 0) {
	do_drop(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (stricmp(cmd, "ACCESS") == 0) {
	do_access(source);

    } else if (stricmp(cmd, "GHOST") == 0) {
	do_ghost(source);

    } else if (stricmp(cmd, "INFO") == 0) {
	do_info(source);

    } else if (stricmp(cmd, "ACC") == 0) {
        do_acc(source);

    } else if (stricmp(cmd, "RECOVER") == 0) {
	do_recover(source);

    } else if (stricmp(cmd, "RELEASE") == 0) {
	do_release(source);

    } else if (stricmp(cmd, "LISTCHANS") == 0) {

	if (!ni) {
		notice(s_NickServ, source, "Only registered nick may use this option.");	
	} else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	        notice(s_NickServ, source, "Please identify with %s first, using the command:", s_NickServ);
	        notice(s_NickServ, source, "/msg %s IDENTIFY \37password\37", s_NickServ);
	} else {
		if (ALLOWCLIST == 1) {
			do_glist(source);
	         } else {
			if (is_services_helpop(source)) {
				do_glist(source);
			} else {
				notice(s_NickServ, source, "This option is disabled by Services Master.");
			} 
		}
	}

    } else if (!is_services_helpop(source) && !is_oper(source)) {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);
        return;

    } else if (stricmp(cmd, "OHELP") == 0) {
	do_ohelp(source);

    } else if (stricmp(cmd, "SENDPASS") == 0) {
        do_sendpass(source);

    } else if (!is_oper(source)) {
         notice(s_NickServ, source, "Permission Denied");
        return;

    } else if (stricmp(cmd, "FLIST") == 0) {
        do_flist(source);

    } else if (!is_services_oper(source)) {
         notice(s_NickServ, source, "Permission Denied");

    } else if (stricmp(cmd, "FORBID") == 0) {
         do_forbid(source);

    } else if (stricmp(cmd, "SETFLAG") == 0) {
         do_setflags(source);

    } else if (!is_services_admin(source)) {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);

    } else if (stricmp(cmd, "GETPASS") == 0) {
	do_getpass(source);

    } else if (stricmp(cmd, "MARK") == 0) {
        char *nick = strtok(NULL, " ");
        NickInfo *ni;
        User *u = finduser(source);

        if ((!nick) || (stricmp(nick, "-") == 0)) {
            notice(s_NickServ,source,"Syntax: \2MARK\2 [-]<nick>");
            notice(s_NickServ,source,
		"\2/msg %s OHELP MARK\2 for more information",s_NickServ);
            return;
        }
        if (*nick == '-') {
           *nick++;
           do_unmark(source, nick);

	} else if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, "Nickname %s is not registered.",nick);
        } else if (!(ni->flags & NI_MARK)) {
            ni->flags |= NI_MARK;
            if (ni->mark)
                  free(ni->mark);
            ni->mark = sstrdup(u->nick);
            slog("NS +M %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            do_break_log("NS_F", "NS +M %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
                "%s set the MARK flag for \2%s\2",u->nick,nick);
        } else {
            notice(s_NickServ,source,"Nickname %s already marked.",nick);
        }

    } else if (stricmp(cmd, "HOLD") == 0) {
        char *nick = strtok(NULL, " ");
        NickInfo *ni;
        User *u = finduser(source);

        if ((!nick) || (stricmp(nick, "-") == 0)) {
            notice(s_NickServ,source,"Syntax: \2HOLD\2 [-]<nick>");
            notice(s_NickServ,source,
                "\2/msg %s OHELP HOLD\2 for more information",s_NickServ);
            return;
        }
        if (*nick == '-') {
           *nick++;
           do_unhold(source, nick);

        } else if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, "Nickname %s is not registered.",nick);
        } else if (!(ni->flags & NI_HOLD)) {
            ni->flags |= NI_HOLD;
            if (ni->hold)
                  free(ni->hold);
            ni->hold = sstrdup(u->nick);
            slog("NS +H %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            do_break_log("NS_F", "NS +H %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
                "%s set the HELD flag for \2%s\2",u->nick,nick);
        } else {
            notice(s_NickServ,source,"Nickname %s already held.",nick);
        }

    } else if (!is_services_root(source)) {
            notice(s_NickServ,source,"Permission Denied.");
            return;

    } else if (stricmp(cmd, "DELETE") == 0) {
        do_delete(source);

    } else if (stricmp(cmd, "SETPASS") == 0) {
        do_setpass(source);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else {
        User *u = finduser(source);
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);

        slog("NS *C %s (%s@%s) [%s]",
            source, u->username, u->host, cmd);
    }

#endif	/* SKELETON */

}

/*************************************************************************/


/* Load/save data files. */

void load_ns_dbase(void)
{
    FILE *f;
    int ver, i, j, len;
    NickInfo *ni;

    if (!(f = open_db(s_NickServ, NICKSERV_DB, "r")))
	return;

    switch (ver = get_file_version(f, NICKSERV_DB)) {
      case 5:
        for (i = 33; i < 256; ++i) {
            while (fgetc(f) == 1) {
                ni = smalloc(sizeof(NickInfo));
                if (1 != fread(ni, sizeof(NickInfo), 1, f))
                    fatal_perror("Read error on %s", NICKSERV_DB);
#ifdef USE_ENCRYPTION
                if (!(ni->flags & (NI_ENCRYPTEDPW | NI_VERBOTEN))) {
                    if (debug)
                        log("debug: %s: encrypting password for `%s' on load",
                                s_NickServ, ni->nick);
                    if (encrypt_in_place(ni->pass, PASSMAX) < 0)
                        fatal("%s: Can't encrypt `%s' nickname password!",
                                s_NickServ, ni->nick);
                    ni->flags |= NI_ENCRYPTEDPW;
                }
#else
                if (ni->flags & NI_ENCRYPTEDPW) {
                    /* Bail: it makes no sense to continue with encrypted
                    * passwords, since we won't be able to verify them */
                    fatal("%s: load database: password for %s encrypted "
                          "but encryption disabled, aborting",
                          s_NickServ, ni->nick);
                }
#endif
                ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
                alpha_insert_nick(ni);
                if (ni->url)
                    ni->url = read_string(f, NICKSERV_DB);
                if (ni->email)
                    ni->email = read_string(f, NICKSERV_DB);
		if (ni->forward)
		    ni->forward = read_string(f, NICKSERV_DB);
                if (ni->hold)
                    ni->hold = read_string(f, NICKSERV_DB);
                if (ni->mark)
                    ni->mark = read_string(f, NICKSERV_DB);
                if (ni->forbid)
                    ni->forbid = read_string(f, NICKSERV_DB);
                ni->last_usermask = read_string(f, NICKSERV_DB);
                ni->last_realname = read_string(f, NICKSERV_DB);
                if (ni->accesscount) {
                    char **access;
                    access = smalloc(sizeof(char *) * ni->accesscount);
                    ni->access = access;
                    for (j = 0; j < ni->accesscount; j++, access++)
                        *access = read_string(f, NICKSERV_DB);
                }
                ni->id_timestamp = 0;
                if (ver < 3) {
                    ni->memomax = DEF_MAX_MEMOS;
                    ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
                } else if (ver == 3) {
                    if (!(ni->flags & (NI_MEMO_SIGNON | NI_MEMO_RECEIVE |
                        NI_NOOP | NI_NOMEMO)))
                        ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE |
                        NI_NOOP | NI_NOMEMO;
                }
            } /* while (fgetc(f) == 1) */
        } /* for (i) */
        break;

      case 4:
      case 3:
      case 2:
      case 1:

	for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {
		ni = smalloc(sizeof(NickInfo));
		if (1 != fread(ni, sizeof(NickInfo), 1, f))
		    fatal_perror("Read error on %s", NICKSERV_DB);
#ifdef USE_ENCRYPTION
		if (!(ni->flags & (NI_ENCRYPTEDPW | NI_VERBOTEN))) {
		    if (debug)
			log("debug: %s: encrypting password for `%s' on load",
				s_NickServ, ni->nick);
		    if (encrypt_in_place(ni->pass, PASSMAX) < 0)
			fatal("%s: Can't encrypt `%s' nickname password!",
				s_NickServ, ni->nick);
		    ni->flags |= NI_ENCRYPTEDPW;
		}
#else
		if (ni->flags & NI_ENCRYPTEDPW) {
		    /* Bail: it makes no sense to continue with encrypted
		    * passwords, since we won't be able to verify them */
		    fatal("%s: load database: password for %s encrypted "
		          "but encryption disabled, aborting",
		          s_NickServ, ni->nick);
		}
#endif
		ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
		alpha_insert_nick(ni);
		if (ni->url)
		    ni->url = read_string(f, NICKSERV_DB);
		if (ni->email)
		    ni->email = read_string(f, NICKSERV_DB);
		ni->last_usermask = read_string(f, NICKSERV_DB);
		ni->last_realname = read_string(f, NICKSERV_DB);
		if (ni->accesscount) {
		    char **access;
		    access = smalloc(sizeof(char *) * ni->accesscount);
		    ni->access = access;
		    for (j = 0; j < ni->accesscount; j++, access++)
			*access = read_string(f, NICKSERV_DB);
		}
		ni->id_timestamp = 0;
		if (ver < 3) {
		    ni->memomax = DEF_MAX_MEMOS;
		    ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
		} else if (ver == 3) {
                    if (!(ni->flags & (NI_MEMO_SIGNON | NI_MEMO_RECEIVE |
                        NI_NOOP | NI_NOMEMO)))
                        ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE |
                        NI_NOOP | NI_NOMEMO;

		}
	    } /* while (fgetc(f) == 1) */
	} /* for (i) */
	break;

      default:
	fatal("Unsupported version number (%d) on %s", i, NICKSERV_DB);
    } /* switch (version) */

    close_db(f, NICKSERV_DB);
}

#ifndef SKELETON


void save_ns_dbase(void)
{
    FILE *f;
    int i, j, len;
    NickInfo *ni;
    char **access;

    if (!(f = open_db(s_NickServ, NICKSERV_DB, "w")))
	return;
    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    fputc(1, f);
	    if (1 != fwrite(ni, sizeof(NickInfo), 1, f))
		fatal_perror("Write error on %s", NICKSERV_DB);
	    if (ni->url)
		write_string(ni->url, f, NICKSERV_DB);
	    if (ni->email)
		write_string(ni->email, f, NICKSERV_DB);
	    if (ni->forward)
		write_string(ni->forward, f, NICKSERV_DB);
            if (ni->hold)
                write_string(ni->hold, f, NICKSERV_DB);
            if (ni->mark)
                write_string(ni->mark, f, NICKSERV_DB);
            if (ni->forbid)
                write_string(ni->forbid, f, NICKSERV_DB);

	    write_string(ni->last_usermask ? ni->last_usermask : "",
							f, NICKSERV_DB);
	    write_string(ni->last_realname ? ni->last_realname : "",
							f, NICKSERV_DB);
	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		write_string(*access, f, NICKSERV_DB);
	}
	fputc(0, f);
    }
    close_db(f, NICKSERV_DB);
}

/*************************************************************************/

/* Install any timeouts we need at startup. */

void ns_install_timeouts(void)
{
#if NS_LAST_SEEN_DELAY > 0
    add_timeout(NS_LAST_SEEN_DELAY, update_last_seen_time, 1);
#endif
}

/*************************************************************************/

/* Check whether a user is on the access list of the nick they're using.
 * If not, send warnings as appropriate.  If so (and not NI_SECURE), update
 * last seen info.  Return 1 if the user is valid and recognized, 0
 * otherwise (note that this means an NI_SECURE nick will always return 0
 * from here). */

int validate_user(User *u)
{
    NickInfo *ni;
    int on_access;

    if (!(ni = findnick(u->nick)))
	return;

    if (ni->flags & NI_VERBOTEN) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick,
		"This nickname may not be used.  Please choose another one.");
	notice(s_NickServ, u->nick,
		"If you do not change within one minute, your nick will "
		"be changed.");
        add_ns_timeout(ni, TO_COLLIDE_TL, 30);
//        add_ns_timeout(ni, TO_COLLIDE, 60);
	return 0;
    }

    if (ni->id_timestamp != 0 && u->signon == ni->id_timestamp)
	;

    on_access = is_on_access(u, ni);

    if (!(ni->flags & NI_SECURE) &&
		(on_access || is_on_id_list(u->nick, u->nick))) {
	ni->flags |= NI_RECOGNIZED;
	if (is_on_id_list(u->nick, u->nick)) {
	    ni->flags |= NI_IDENTIFIED;
    	    ni->last_seen = time(NULL);
	    if (ni->last_usermask)
	        free(ni->last_usermask);
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    if (ni->last_realname)
 	        free(ni->last_realname);
	    ni->last_realname = sstrdup(u->realname);
        }
	return 1;
    }

    if (ni->flags & NI_SECURE) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick,
		"This nickname is registered and protected.  If it is your");
	notice(s_NickServ, u->nick,
		"nick, type \2/msg %s IDENTIFY \37password\37\2.  Otherwise,",
		s_NickServ);
	notice(s_NickServ, u->nick,
		"please choose a different nick.");
    } else {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick,
		"This nick is owned by someone else.  Please choose another.");
	notice(s_NickServ, u->nick,
		"(If this is your nick, type \2/msg %s IDENTIFY \37password\37\2.)",
		s_NickServ);
    }

    if ((ni->flags & NI_KILLPROTECT)
		&& !((ni->flags & NI_SECURE) && (on_access || is_on_id_list(u->nick, u->nick)))) {
	notice(s_NickServ, u->nick,
		"If you do not change within one minute, your nick will be "
		"changed.");
        add_ns_timeout(ni, TO_COLLIDE_TL, 30);
//        add_ns_timeout(ni, TO_COLLIDE, 60);

    }

    return 0;
}

/*************************************************************************/

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickInfo *ni;

    if (ni = findnick(u->nick)) {
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	del_ns_timeout(ni, TO_COLLIDE);
    }
}

/*************************************************************************/

/* Remove all nicks which have expired. */

void expire_nicks()
{
    NickInfo *ni, *next;
    int i;
    const time_t expire_time = NICK_EXPIRE*24*60*60;
    time_t now = time(NULL);
    long count = 0;
    long xcount = 0;

    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = next) {
	    next = ni->next;
	    if (now - ni->last_seen >= expire_time
					&& !(ni->flags & NI_VERBOTEN)
					&& !(ni->flags & NI_HOLD)) {
		++xcount;
                slog("NS X %s", ni->nick);
                do_break_log("NS_X", "NS X %s", ni->nick);

		log("Expiring nickname %s", ni->nick);
		delnick(ni);
	    }
	}
    }
    for (i = 33; i < 256; ++i) {
        for (ni = nicklists[i]; ni; ni = ni->next)
            ++count;
    }
    if (DISPLAY_UPDATES == 1)
        wallops(SERVER_NAME, "Completed NickName Expire (%d/%d)",
        	xcount,count);
}

#endif /* Skeleton */

/*************************************************************************/

/* Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */

NickInfo *findnick(const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next) {
	if (stricmp(ni->nick, nick) == 0)
	    return ni;
    }
    return NULL;
}

#ifndef SKELETON

/*************************************************************************/
/*********************** NickServ private routines ***********************/
/*************************************************************************/

/* Is the given user's address on the given nick's access list?  Return 1
 * if so, 0 if not. */

int is_on_access(User *u, NickInfo *ni)
{
    int i;
    char *buf;

    i = strlen(u->username);
    buf = smalloc(i + strlen(u->host) + 2);
    sprintf(buf, "%s@%s", u->username, u->host);
    strlower(buf+i+1);
    for (i = 0; i < ni->accesscount; ++i) {
	if (match_wild(ni->access[i], buf)) {
	    free(buf);
	    return 1;
	}
    }
    free(buf);
    return 0;
}

/*************************************************************************/

#endif /* SKELETON */

/* Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists[tolower(*nick)];
			ni2 && stricmp(ni2->nick, nick) < 0;
			ni3 = ni2, ni2 = ni2->next)
	;
    ni->prev = ni3;
    ni->next = ni2;
    if (!ni3)
	nicklists[tolower(*nick)] = ni;
    else
	ni3->next = ni;
    if (ni2)
	ni2->prev = ni;
}

/*************************************************************************/

#ifndef SKELETON

/* Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */

static NickInfo *makenick(const char *nick)
{
    NickInfo *ni;
    User *u;

    ni = scalloc(sizeof(NickInfo), 1);
    strscpy(ni->nick, nick, NICKMAX);
    alpha_insert_nick(ni);
    return ni;
}

/*************************************************************************/

/* Remove a nick from the NickServ database.  Return 1 on success, 0
 * otherwise.  Also deletes the nick from any channel access lists it is
 * on. */

static int delnick(NickInfo *ni)
{
    int i;

    if (ni->next)
	ni->next->prev = ni->prev;
    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklists[tolower(*ni->nick)] = ni->next;
    if (ni->last_usermask)
	free(ni->last_usermask);
    if (ni->last_realname)
	free(ni->last_realname);
    if (ni->access) {
	for (i = 0; i < ni->accesscount; ++i)
	    free(ni->access[i]);
	free(ni->access);
    }
    if (ni->hold)
      free(ni->hold);
    if (ni->mark);
      free(ni->mark);
    if (ni->forbid)
      free(ni->forbid);
    cs_remove_nick(ni->nick);
    clear_memos(ni->nick);
    os_remove_nick(ni->nick);
    free(ni);
    return 1;
}

/*************************************************************************/

/* Collide a nick. */

static void collide(NickInfo *ni, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(ni, TO_COLLIDE);
    
    send_cmd(NULL, "SVSNICK %s %s%d 0", ni->nick, "Guest", getrandom(10000, 300000));
    send_cmd(NULL, "NICK %s 1 %lu %s enforcer %s %s 0 :Nick Protection Enforcement",
           ni->nick, time(NULL), "+", services_host, server_name);
    ni->flags |= NI_KILL_HELD;
    add_ns_timeout(ni, TO_RELEASE, RELEASE_TIMEOUT);
}

/*************************************************************************/

/* Release hold on a nick. */

static void release(NickInfo *ni, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(ni, TO_RELEASE);
    send_cmd(ni->nick, "QUIT");
    ni->flags &= ~NI_KILL_HELD;
}

/*************************************************************************/
/*************************************************************************/

static struct my_timeout {
    struct my_timeout *next, *prev;
    NickInfo *ni;
    Timeout *to;
    int type;
} *my_timeouts;

/*************************************************************************/

/* Remove a collide/release timeout from our private list. */

static void rem_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}

/*************************************************************************/

/* Collide a nick on timeout. */

static void timeout_collide(Timeout *t)
{
    NickInfo *ni = t->data;
    User *u;

    rem_ns_timeout(ni, TO_COLLIDE);
    /* If they identified or don't exist anymore, don't kill them. */
    if (ni->flags & NI_IDENTIFIED
		|| !(u = finduser(ni->nick))
		|| u->my_signon > t->settime)
	return;
    /* The RELEASE timeout will always add to the beginning of the
     * list, so we won't see it.  Which is fine because it can't be
     * triggered yet anyway. */
    collide(ni, 1);
}

static void timeout_collide_tl(Timeout *t)
{
   NickInfo *ni = t->data;
   User *u;

   rem_ns_timeout(ni, TO_COLLIDE_TL);
   if (ni->flags & NI_IDENTIFIED
		|| !(u = finduser(ni->nick))
		|| u->my_signon > t->settime)
	return;
   add_ns_timeout(ni, TO_COLLIDE, 30);
   notice(s_NickServ, ni->nick,
          "If you do not change within 30 seconds, your nick will "
          "be changed.");
}

/*************************************************************************/

/* Release a nick on timeout. */

static void timeout_release(Timeout *t)
{
    NickInfo *ni = t->data;

    rem_ns_timeout(ni, TO_RELEASE);
    release(ni, 1);
}

/*************************************************************************/

/* Add a collide/release timeout. */

static void add_ns_timeout(NickInfo *ni, int type, time_t delay)
{
    Timeout *to;

    struct my_timeout *t;
    void (*timeout_routine)(Timeout *);

    if (type == TO_COLLIDE_TL)
        timeout_routine = timeout_collide_tl;
    else if (type == TO_COLLIDE)
	timeout_routine = timeout_collide;
    else if (type == TO_RELEASE)
	timeout_routine = timeout_release;
    else {
	log("NickServ: unknown timeout type %d!  ni=%p (%s), delay=%ld",
		type, ni, ni->nick, delay);
	return;
    }

    to = add_timeout(delay, timeout_routine, 0);
    to->data = ni;
    t = smalloc(sizeof(*t));
    t->next = my_timeouts;
    my_timeouts = t;
    t->prev = NULL;
    t->ni = ni;
    t->to = to;
    t->type = type;
}

/*************************************************************************/

/* Delete a collide/release timeout. */

static void del_ns_timeout(NickInfo *ni, int type)
{

    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    del_timeout(t->to);
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }  
}

/*************************************************************************/
/*************************************************************************/

/* Update last-seen time for all users currently online. */

static void update_last_seen_time(Timeout *to_unused)
{
    User *u;
    NickInfo *ni;

    /* This is O(n^2) at the moment, because the user list isn't known to
     * be sorted.
     */
    for (u = first_user(); u; u = u->next) {
	if (ni = findnick(u->nick))
            if (ni && (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)))
        	 ni->last_seen = time(NULL);
    }
}

/*************************************************************************/
/*********************** NickServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_ohelp(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];
    int is_servadmin = is_services_oper(source);

    if (!cmd)
        slog("NS - %s requested ohelp: (index)", source);

    if (cmd) {
        slog("NS - %s requested ohelp: %s", source, cmd);

        if (stricmp(cmd, "DELETE") == 0) {
            notice_list(s_NickServ, source, oper_delete_help);
            return;
        } else if (stricmp(cmd, "LIST") == 0) {
            notice_list(s_NickServ, source, oper_list_help);
            return;
        } else if (stricmp(cmd, "FLIST") == 0) {
            notice_list(s_NickServ, source, flist_help);
            return;
        } else if (stricmp(cmd, "GETPASS") == 0) {
            notice_list(s_NickServ, source, getpass_help);
            return;
        } else if (stricmp(cmd, "FORBID") == 0) {
            notice_list(s_NickServ, source, forbid_help);
            return;
        } else if (stricmp(cmd, "SENDPASS") == 0) {
            notice_list(s_NickServ, source, sendpass_help);
            return;
        } else if (stricmp(cmd, "MARK") == 0) {
            notice_list(s_NickServ, source, mark_help);
            return;
        } else if (stricmp(cmd, "HOLD") == 0) {
            notice_list(s_NickServ, source, hold_help);
            return;
        } else if (stricmp(cmd, "SETPASS") == 0) {
            notice_list(s_NickServ, source, setpass_help);
            return;
      
        } else if (stricmp(cmd, "SETFLAG") == 0) {
            notice_list(s_NickServ, source, setflags_help);
            return;
        }

    }

    if (!cmd)
        notice_list(s_NickServ, source, ns_servadmin_help);

    notice(s_NickServ, source,
       "\2END OF OHELP\2");

}

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];

    snprintf(buf, sizeof(buf), "%s%s", s_NickServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_NickServ, source, buf);
}

/*************************************************************************/

/* Register a nick. */

static void do_register(const char *source)
{
    NickInfo *ni;
    User *u;
    char *gender;
    char *pass = strtok(NULL, " ");
    char *email = strtok(NULL, " ");
    int i;

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname registration is temporarily disabled.");
	return;
    }

    if (!pass || !email) {

        notice(s_NickServ, source, "Syntax: \2REGISTER\2 <password> <email>"); 
	notice(s_NickServ, source,
			"\2/msg %s HELP REGISTER\2 for more information.",
			s_NickServ);

    } else if (!(u = finduser(source))) {

	log("%s: Can't register nick %s: nick not online", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, registration failed.");

    } else if (time(NULL) < u->lastnickreg + NS_REG_DELAY) {

	notice(s_NickServ, source,
	    "Please wait %d seconds before using the REGISTER command again.",
	    NS_REG_DELAY);

    } else if (ni = findnick(source)) {

	if (ni->flags & NI_VERBOTEN) {
	    log("%s: %s@%s tried to register FORBIDden nick %s", s_NickServ,
			u->username, u->host, source);
	    notice(s_NickServ, source,
			"Nickname \2%s\2 may not be registered.", source);
	} else {
	    notice(s_NickServ, source,
			"Nickname \2%s\2 is already registered!", source);
	}

    } else if (stricmp(source, pass) == 0 || strlen(pass) < 5) {

	notice(s_NickServ, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);

     } else if (!strchr(email, '@')) {
            notice(s_NickServ, source,
               "ERROR: E-Mail Address must contain an '@'");
     } else if (!strchr(email, '.')) {
            notice(s_NickServ, source,
               "ERROR: Look like you give invalid email.");
     } else if (strchr(email, '\\')) {
           notice(s_NickServ, source,
               "ERROR: invalid email characters.");
    } else {

	if (ni = makenick(source)) {
#ifdef USE_ENCRYPTION
	    int len = strlen(pass);
	    if (len > PASSMAX)
		len = PASSMAX;
	    if (encrypt(pass, len+1, ni->pass, PASSMAX) < 0) {
		memset(pass, 0, strlen(pass));
		log("%s: Failed to encrypt password for %s (register)",
			s_NickServ, source);
		notice(s_NickServ, source, "Sorry, registration failed.");
		return;
	    }
	    memset(pass, 0, strlen(pass));
	    ni->flags = NI_ENCRYPTEDPW | NI_IDENTIFIED | NI_RECOGNIZED;
            if (!is_on_id_list(source, source)) {
                for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                    ;
                if (i < MAX_IDS)
                     u->id_nicks[i] = sstrdup(source);
            }
#else
	    strscpy(ni->pass, pass, PASSMAX);
	    ni->flags = NI_IDENTIFIED | NI_RECOGNIZED;
            if (!is_on_id_list(source, source)) {
                for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                    ;
                if (i < MAX_IDS)
                     u->id_nicks[i] = sstrdup(source);
            }
#endif

	    ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
            if (AUTO_NOHOST > 0)
                ni->flags |= NI_NOHOST; 
	    if (NS_PRIVATE > 0)
	   	ni->flags |= NI_PRIVATE;
	    if (NS_KILL > 0)
		ni->flags |= NI_KILLPROTECT;
	    if (NS_SECURE > 0)
		ni->flags |= NI_SECURE;

	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    ni->last_realname = sstrdup(u->realname);
	    ni->time_registered = ni->last_seen = time(NULL);
	    ni->email = sstrdup (email);
	    ni->accesscount = 1;
	    ni->access = smalloc(sizeof(char *));
	    ni->access[0] = create_mask(u);
            ni->memomax = DEF_MAX_MEMOS;
	    clear_memos(ni->nick);
	    log("%s: `%s' registered by %s@%s", s_NickServ,
		    source, u->username, u->host);
            slog("NS R %s (%s@%s)", source, u->username, u->host);
            do_break_log("NS_R", "NS R %s (%s@%s)",
                 source, u->username, u->host);

            notice(s_NickServ, source, NS_REGISTER, source, ni->access[0], email);
#ifndef USE_ENCRYPTION
            notice(s_NickServ, source, NS_REGISTER2, pass);
 	    notice(s_NickServ, source, "If you wish to keep your E-Mail secret, type \2/MSG %s SET HideMail ON\2.", s_NickServ);
	    wallops(s_NickServ, "Nick \2%s\2 is registered to %s",u->nick, ni->access[0]);

#endif
#if REG_SAVES > 0
            if (regcnt++ >= REG_SAVES) {
                wallops(SERVER_NAME,
                     "Automatic DataBase save on %d new registrations",
                          regcnt);
                save_data = 1;
            }
#endif

            show_next_db(source, s_NickServ);
	    u->lastnickreg = time(NULL);
               send_cmd(s_NickServ, "SVSMODE %s +r 0", source);
	} else {
	    log("%s: makenick(%s) failed", s_NickServ, source);
	    notice(s_NickServ, source,
		"Sorry, couldn't register your nickname.");
	}

    }

}

/*************************************************************************/
static void do_identify(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    char *tmp;
    NickInfo *ni;
    User *u;
    int res;
    char access[128];
//    char **link;
    int i;
    int j;
    int is_force;

    *access = 0;

    if (!nick) {

        notice(s_NickServ, source, "Syntax: \2IDENTIFY\2 [nick] <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.",
		s_NickServ);
	return;
    }
    if (!pass) {
	pass = sstrdup(nick);
	nick = sstrdup(source);
    }
    ni = findnick(nick);

    is_force = (is_services_root(source) && !stricmp(pass, "force"));
    if (!ni) {
	notice(s_NickServ, source, "Your nick (%s) isn't registered.",nick);
    } else if (!(u = finduser(source))) {

	log("%s: IDENTIFY from nonexistent nick %s", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, identification failed.");

    } else if (u->flags & U_ABUSIVE) {

	log("%s: Failed identify from abusive user %s", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, You have the abusive flag set");
      if (EXTRA_SNOOP > 0) {
          slog("NS *I %s (%s!%s@%s) [Abusive]",
		nick, source, u->username, u->host);
          do_break_log("NS_I", "NS *I %s (%s!%s@%s)",
              nick, source, u->username, u->host);
      }


    } else if (!is_force && !(res = check_password(pass, ni->pass))) {

	log("%s: Failed IDENTIFY for %s (%s!%s@%s)",
		s_NickServ, nick, source, u->username, u->host);
        if (EXTRA_SNOOP) {
            slog("NS *I %s (%s!%s@%s)",
	         nick, source, u->username, u->host);
            do_break_log("NS_I", "NS *I %s (%s!%s@%s)",
                 nick, source, u->username, u->host);
        }
        notice(s_NickServ, source, NS_BADPASS);
	bad_password(u, nick);

    } else if (res == -1) {

	notice(s_NickServ, source, "Sorry, identification failed.");

    } else {

        if (is_autokilled(source)) {
          wallops(s_NickServ, "\2Warning\2 - Found identified nick \2%s\2 on AUTOKILL list", source);
          slog("NS A*I %s (%s!%s@%s)",
                source, source, u->username, u->host);
          do_break_log("NS_I", "NS A*I %s (%s!%s@%s)",
               source, source, u->username, u->host);
          kill_user(SERVER_NAME, source, "You are Auto-KILLED from this network, please stay away!");
          return;
       }


      if ((!(ni->flags & NI_IDENTIFIED)) && (EXTRA_SNOOP > 0)) {
          slog("NS I %s (%s!%s@%s) %s",
		nick, source, u->username, u->host,
                is_force ? "[Forced]" : "");
          do_break_log("NS_I", "NS I %s (%s!%s@%s) %s",
		nick, source, u->username, u->host,
                is_force ? "[Forced]" : "");
       }

     if (is_hostline(source))
	do_hostline(source);

     if (ni->flags & NI_IDENTIFIED) {
        notice(s_NickServ, source, "You are already identify.");
	do_doop(source);
	return;	
     }

	ni->flags |= NI_IDENTIFIED;
	ni->id_timestamp = u->signon;


	if (!is_on_id_list(u->nick, nick)) {
	    for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
		;
	    if (i < MAX_IDS)
	         u->id_nicks[i] = sstrdup(nick);
	}

/*        for (j=0; ; ++link, ++j) {
            if (!is_on_id_list(u->nick, *link)) {
                for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                    ;
        	if (i < MAX_IDS)
     	            u->id_nicks[i] = sstrdup(*link);
	    }
	}
*/
	if (!(ni->flags & NI_RECOGNIZED)) {
	    ni->last_seen = time(NULL);
	    if (ni->last_usermask)
		free(ni->last_usermask);
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    if (ni->last_realname)
		free(ni->last_realname);
	    ni->last_realname = sstrdup(u->realname);
	    log("%s: %s!%s@%s identified for nick %s %s", s_NickServ,
			source, u->username, u->host, source,
                        is_force ? "[Forced]" : "");
	}
	if (is_services_coder(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aq 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Master",u->nick);
                strcpy(access, "as a Services Master.");
	} else if (is_services_root(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aq 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Root",u->nick);
                strcpy(access, "as a Services Root.");
	} else if (is_services_admin(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aq 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Admin",u->nick);
//                strcpy(access, "as a Services Admin.");
	} else if (is_services_oper(source)) {
		wallops(NULL,"%s now is a Services IRCop", u->nick);
                send_cmd(s_NickServ,"SVSMODE %s +ohwefbcgswW 0",source);
                strcpy(access, "as a Services Operator.");
// 	        send_cmd(s_NickServ,"SWHOIS %s :is a Services Operator.",u->nick);
	} else if (is_services_helpop(source)) {
		wallops(NULL,"%s now is a Services Local IRCop", u->nick);
                send_cmd(s_NickServ,"SVSMODE %s +OhwefbcgswW 0",source);
                strcpy(access, "as a Services HelpOp.");
//	        send_cmd(s_NickServ,"SWHOIS %s :is a Services HelpOp.",u->nick);
	} else {
                strcpy(access, "");
	}
	do_doop(source);

        notice(s_NickServ, source, NS_IDENTIFY,ni->nick,access);
            send_cmd(s_NickServ, "SVSMODE %s +r 0", source);

	if (!(ni->flags & NI_RECOGNIZED))
	    check_memos(source);
    }
}

/*************************************************************************/

static void do_delete(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);
    int is_root = is_services_root(source);

    if (readonly) {
	notice(s_NickServ, source,
                "Sorry, nickname deletion is temporarily disabled. (Read Only)");
	return;
    }
    if (!nick) {
       notice(s_NickServ, source, "\2Delete\2: You MUST specify a nick");
       return;
    }

    if (!(ni = findnick(nick))) {
        if (nick)
           notice(s_NickServ, source, "Nick %s isn't registered.", nick);
        return;
    }
    if (stricmp(nick, SERVICES_MASTER) == 0 && !is_services_coder(source)) {
        notice(s_NickServ, source,
            "You are not permitted to delete a Master's account");
        return;
    }

    delnick(ni);
    slog("NS +D %s (%s!%s@%s)", nick, source, u->username, u->host);
    do_break_log("NS_G", "NS +D %s (%s!%s@%s)",
       nick, source, u->username, u->host);
    log("%s: %s!%s@%s deleted nickname %s", s_NickServ,
            source, u->username, u->host, nick);
        notice(s_NickServ, source, NS_DELETE, nick);
        wallops(s_NickServ, "%s deleted nickname \2%s\2", source, nick);
}



/*************************************************************************/

static void do_drop(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname de-registration is temporarily disabled.");
	return;
    }

    if (!nick) {
       notice(s_NickServ, source, "Syntax: \2DROP\2 <nick>");
       return;
    }

    if (!(ni = findnick(nick))) {
	    notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!is_on_id_list(source, nick)) {

        slog("NS *D %s (%s!%s@%s) [User Hasn't Identified]", source, source, u->username, u->host);
        do_break_log("NS_O", "NS *D %s (%s!%s@%s) [User Hasn't Identified]", source, source, u->username, u->host);
	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY [\37nick\37] \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: DROP from nonexistent user %s", s_NickServ, source);

    } else if (!u) {

	log("%s: DROP %s from nonexistent oper %s", s_NickServ, nick, source);
	notice(s_NickServ, source, "Can't find your user record!");

    } else {

	if (readonly) {
	    notice(s_NickServ, source,
		"Warning: Services is in read-only mode.  Changes will not be saved.");
	}
        if (stricmp(nick, SERVICES_MASTER) == 0 && !is_services_coder(source)) {
            notice(s_NickServ, source,
               "You are not permitted to delete a Master's account");
            return;
        }
	delnick(ni);
        slog("NS D %s (%s!%s@%s)", nick, source, u->username, u->host);
        do_break_log("NS_O", "NS D %s (%s!%s@%s)",
             nick, source, u->username, u->host);
	log("%s: %s!%s@%s dropped nickname %s", s_NickServ,
		source, u->username, u->host, nick ? nick : source);
        notice(s_NickServ, source, NS_DROP);
    }
}

/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname option setting is temporarily disabled.");
	return;
    }

    if (!param) {

	notice(s_NickServ, source,
                "Syntax: \2SET\2 <command> [<parameters>]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nickname is not registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "PASSWORD") == 0) {

	do_set_password(ni, param);

    } else if (stricmp(cmd, "URL") == 0) {

	do_set_url(ni, param);

    } else if (stricmp(cmd, "EMAIL") == 0) {

	do_set_email(ni, param);

    } else if (stricmp(cmd, "KILL") == 0) {

	do_set_kill(ni, param);

    } else if (stricmp(cmd, "SECURE") == 0) {

	do_set_secure(ni, param);

    } else if (stricmp(cmd, "PRIVATE") == 0) {

	do_set_private(ni, param);

    } else if (stricmp(cmd, "EMAILMEMOS") == 0) {
	do_set_emailmemos(ni, param);

    } else if (stricmp(cmd, "NEVEROP") == 0) {
        do_set_neverop(ni, param);

    } else if (stricmp(cmd, "NOOP") == 0) {
        do_set_noop(ni, param);

    } else if (stricmp(cmd, "AUTH") == 0) {
        do_set_auth(ni, param);

    } else if (stricmp(cmd, "MALAY") == 0) {
        do_set_malay(ni, param);

    } else if (stricmp(cmd, "PRIVMSG") == 0) {
        do_set_privmsg(ni, param);

    } else if (stricmp(cmd, "NOMEMO") == 0) {
        do_set_nomemo(ni, param);

    } else if (stricmp(cmd, "NOHOST") == 0) {
        do_set_nohost(ni, param);

    } else if (stricmp(cmd, "HIDEMAIL") == 0) {
	do_set_hidemail(ni, param);

    } else {

	notice(s_NickServ, source,
			"Unknown SET option \2%s\2.", strupper(cmd));

    }
}

/*************************************************************************/

static void do_set_password(NickInfo *ni, char *param)
{
    char *source = ni->nick;
#ifdef USE_ENCRYPTION
    int len = strlen(param);
    if (len > PASSMAX)
	len = PASSMAX;
    if (encrypt(param, len, ni->pass, PASSMAX) < 0) {
	memset(param, 0, strlen(param));
	log("%s: Failed to encrypt password for %s (set)",
		s_NickServ, ni->nick);
	notice(s_NickServ, ni->nick, "Sorry, couldn't change password.");
	return;
    }
    memset(param, 0, strlen(param));
    notice(s_NickServ, ni->nick, "Password changed.");
#else
    strscpy(ni->pass, param, PASSMAX);
    notice(s_NickServ, ni->nick, "Password changed to \2%s\2.", param);
    show_next_db(source, s_NickServ);
#endif
}

/*************************************************************************/

static void do_set_url(NickInfo *ni, char *param)
{
    char *source = ni->nick;
    if (ni->url)
	free(ni->url);

    if (stricmp(param,"none")) {
        ni->url = sstrdup(param);
        notice(s_NickServ, ni->nick,
		"URL for %s changed to \2%s\2.", ni->nick, param);
    } else {
       if (!ni->url) {
            notice(s_NickServ, ni->nick,
                "You have no url set to disable!");
            return;
       }

       ni->url = NULL;
       notice(s_NickServ, source, "URL disabled for nick \2%s\2", ni->nick);
    }
    show_next_db(source, s_NickServ);
}

/*************************************************************************/

static void do_set_email(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if ((stricmp(param,"none")) && (stricmp(param,"hide"))) {

       if (!strchr(param, '@')) {
            notice(s_NickServ, source,
               "ERROR: E-Mail Address must contain an '@'");
            return;
       }
       if (strchr(param, '\\')) {
           notice(s_NickServ, source,
               "ERROR: invalid email characters.");
           return;
       }

       if (ni->email)
           ni->email = NULL;

       ni->email = sstrdup(param);
       notice(s_NickServ, ni->nick,
		"E-mail address for %s changed to \2%s\2.", ni->nick, param);
       slog("NS E %s [%s]", ni->nick, param);

    } else if (!stricmp(param,"hide")) {
         if (!ni->email) {
            notice(s_NickServ, ni->nick,
                "You have no email address set to disable!");
            return;
         }
         ni->flags |= NI_HIDE_EMAIL;
         notice(s_NickServ, ni->nick,
                  "Your email address will no longer be shown");
         show_next_db(source, s_NickServ);
         return;

    } else {
       if (!ni->email) {
           notice(s_NickServ, ni->nick,
               "You have no email address set to disable!");
           return;
       }

       ni->flags &= ~NI_HIDE_EMAIL;
       ni->email = NULL;
       notice(s_NickServ, source, "EMAIL disabled for nick \2%s\2", ni->nick);
    }
    show_next_db(source, s_NickServ);
}

/*************************************************************************/

static void do_set_kill(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET KILL\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET KILL\2 for more information.", s_NickServ);
    }
}

/*************************************************************************/

static void do_set_secure(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET SECURE\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET SECURE\2 for more information.",
		s_NickServ);
    }
}

/*************************************************************************/

static void do_set_private(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET PRIVATE\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET PRIVATE\2 for more information.",
		s_NickServ);
    }
}

static void do_set_emailmemos(NickInfo *ni, char *param)
{
    char *source = ni->nick;


    if (stricmp(param, "ON") == 0) {
	if (!ni->email) {
	    notice(s_NickServ,source,"You must set your e-mail address.");
	    notice(s_NickServ,source,"Do do this type: \2/msg %s set email \37email\37\2",s_NickServ);
	    return;
        } else if ((ALLOW_EMAIL == 0) && !is_services_root(source)) {
              notice(s_NickServ, source,
                    "Sorry, emailing memos have been disabled");
              return;
        }
        ni->flags |= NI_EMAILMEMOS;
        notice(s_NickServ, source, "Your memos will now be emailed to %s.", ni->email);
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

        ni->flags &= ~NI_EMAILMEMOS;
        notice(s_NickServ, source, "Your memos will no longer be emailed.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET EMAILMEMOS\2 [\37ON\37|\37OFF\37]");
        notice(s_NickServ, source,
                "\2/msg %s HELP SET EMAILMEMOS\2 for more information.",
                s_NickServ);
    }
}


/*************************************************************************/

static void do_set_neverop(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2Never Op\2 setting enabled");
       ni->flags |= NI_NEVEROP;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2Never Op\2 setting disabled");
       ni->flags &= ~NI_NEVEROP;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NEVEROP\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NEVEROP\2 for more information.",
              s_NickServ);

  }
}



/*************************************************************************/

static void do_set_noop(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Op\2 setting enabled");
       ni->flags |= NI_NOOP;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Op\2 setting disabled");
       ni->flags &= ~NI_NOOP;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOOP\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOOP\2 for more information.",
              s_NickServ);

  }
}

/*************************************************************************/

static void do_set_nomemo(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Memo\2 setting enabled");
       ni->flags |= NI_NOMEMO;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Memo\2 setting disabled");
       ni->flags &= ~NI_NOMEMO;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOMEMO\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOMEMO\2 for more information.",
              s_NickServ);

  }
}

static void do_set_nohost(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Host Display\2 setting enabled");
       ni->flags |= NI_NOHOST;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Host Display\2 setting disabled");
       ni->flags &= ~NI_NOHOST;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOHOST\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOHOST\2 for more information.",
              s_NickServ);
   }
}

/*************************************************************************/
/*
static void do_link(const char *source)
{

    char *cmd = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    int i;
    char **link;
    if ((CANLINK = 0) && !is_services_root(source)){
	notice(s_NickServ, source,
		"Sorry, Nickname link is disable");
	return;
    }
    if ((CANLINK = 2) && !is_services_root(source)) {
	notice(s_NickServ, source,
		"Sorry, Nickname link is disable");
	return;
    }
    if (cmd && stricmp(cmd, "LIST") == 0 && nick && is_services_admin(source)) {
        if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, "Nick \2%s\2 not registered!", nick);
            return;
        }
        notice(s_NickServ, source, "Link list for \2%s\2:", nick);
        for (link = ni->link, i = 0; i < ni->linkcount; ++link, ++i)
            notice(s_NickServ, source, "    %s", *link);
	notice(s_NickServ, source, "\2End of List\2");

    } else if (!cmd || (((stricmp(cmd,"LIST")==0) || (stricmp(cmd,"WIPE")==0)) ? !!nick : !nick)) {
        notice(s_NickServ, source,
                "Syntax: \2LINK {ADD|DEL|LIST|WIPE} [\37mask\37]\2");
        notice(s_NickServ, source,
                "\2/msg %s HELP LINK\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {
       notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

        notice(s_NickServ, source,
                "Password authentication required for that command.");
        notice(s_NickServ, source,
                "Retry after typing \2/msg %s IDENTIFY \37password\37.",
                s_NickServ);
        if (!u)
            log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "ADD") == 0) {

	if (stricmp(source, nick) == 0) {
	    notice(s_NickServ,source, "There is no reason to add your "
		"nick to your own link list.");
	    return;
	}

        for (link = ni->link, i = 0; i < ni->linkcount; ++link, ++i) {
            if (strcmp(*link, nick) == 0) {
                notice(s_NickServ, source,
                        "Nick \2%s\2 already present on your link list.",
                        *link);
                return;
            }
        }
	if (!is_on_id_list(source, nick)) {
	    notice(s_NickServ,source,"Please identify to that nickname "
		"before adding it to your link list.");
	    notice(s_NickServ,source,"/msg %s IDENTIFY %s \37password\37",
		s_NickServ, nick);
	    return;
	}
        ++ni->linkcount;
        ni->link = srealloc(ni->link, sizeof(char *) * ni->linkcount);
        ni->link[ni->linkcount-1] = sstrdup(nick);
        notice(s_NickServ, source, "\2%s\2 added to your link list.", nick);

    } else if (stricmp(cmd, "DEL") == 0) {

        for (link = ni->link, i = 0; i < ni->linkcount; ++link, ++i) {
            if (strcmp(*link, nick) == 0)
                break;
        }
        if (i == ni->linkcount) {
            for (link = ni->link, i = 0; i < ni->linkcount;
                                                        ++link, ++i) {
                if (stricmp(*link, nick) == 0)
                    break;
            }
        }
        if (i == ni->linkcount) {
            notice(s_NickServ, source,
                        "\2%s\2 not found on your link list.", nick);
            return;
        }
        notice(s_NickServ, source,
                "\2%s\2 deleted from your link list.", *link);
        free(*link);
        --ni->linkcount;
        if (i < ni->linkcount)
            bcopy(link+1, link, (ni->linkcount-i) * sizeof(char *));
        if (ni->linkcount)
            ni->link = srealloc(ni->link, ni->linkcount * sizeof(char *));
        else {
            free(ni->link);
            ni->link = NULL;
        }

    } else if (stricmp(cmd, "LIST") == 0) {

        notice(s_NickServ, source, "\2Link list:\2");
        for (link = ni->link, i = 0; i < ni->linkcount; ++link, ++i) {
            notice(s_NickServ, source, "    %s", *link);
        }
        notice(s_NickServ, source, "\2End of List\2");

    } else if (stricmp(cmd, "WIPE") == 0) {
        for (link = ni->link, i = 0; i < ni->linkcount; ++link, ++i) {
              free(*link);
        }
        ni->linkcount=0;
        free(ni->link);
        ni->link = NULL;
        notice(s_NickServ,source,"Your link list has been wiped.");

    } else {

        notice(s_NickServ, source,
                "Syntax: \2LINK {ADD|DEL|LIST|WIPE} [\37mask\37]\2");
        notice(s_NickServ, source,
                "\2/msg %s HELP LINK\2 for more information.", s_NickServ);
    }
}
*/

static void do_access(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    int i;
    char **access;

    if (cmd && stricmp(cmd, "LIST") == 0 && mask && is_services_admin(source)) {

	if (!(ni = findnick(mask))) {
	    notice(s_NickServ, source, "Nick \2%s\2 not registered!", mask);
	    return;
	}
	notice(s_NickServ, source, "Access list for \2%s\2:", mask);
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    notice(s_NickServ, source, "    %s", *access);

    } else if (!cmd || (((stricmp(cmd,"LIST")==0) || (stricmp(cmd,"WIPE")==0)) ? !!mask : !mask)) {

	notice(s_NickServ, source,
                "Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (mask && ((!strchr(mask, '@') || strchr(mask, '!')) &&
               (stricmp(cmd, "ADD") == 0))) {

	notice(s_NickServ, source,
		"Mask must be in the form \37user\37@\37host\37.");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "ADD") == 0) {

       if (strlen(mask) > 60) {
           notice(s_NickServ, source, "NickServ access mask too large!");
           wallops(s_NickServ, "%s tried to add a HUGE mask to the access list", source);
           return;
       }
       if (stricmp(mask, "*@*") == 0 || stricmp(mask, "*@*.*") == 0) {
           notice(s_NickServ, source,
               "Please try to add a more secure access mask");
           return;
        }

	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0) {
		notice(s_NickServ, source,
			"Mask \2%s\2 already present on your access list.",
			*access);
		return;
	    }
	}
        if ((ni->accesscount > USER_ACCESS_MAX-1) && (USER_ACCESS_MAX > 0)) {
            notice(s_NickServ, source,
                 "Sorry, access list full for nick %s", ni->nick);
            wallops(SERVER_NAME, "%s tried to add another access mask", ni->nick);
            return;
        }
            
	++ni->accesscount;
	ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
	ni->access[ni->accesscount-1] = sstrdup(mask);
	notice(s_NickServ, source, "\2%s\2 added to your access list.", mask);
        show_next_db(source, s_NickServ);

        if (readonly)
            notice(s_NickServ, source,
               "Warning: Services is in read-only mode; changes will not be saved.");

    } else if (stricmp(cmd, "DEL") == 0) {
        int z, j=0;

        if (strspn(mask, "1234567890") == strlen(mask) &&
                            (i = atoi(mask)) > 0 && i <= ni->accesscount) {

            for (access = ni->access, z = 0; z < ni->accesscount;
                        ++access, ++z) {

               j += 1;
               if (i == j) {
                   access = &ni->access[z];
                   break;
               }
               
            }
            if (!(i == j)) {
                notice(s_NickServ, source,
                        "No such entry (#%d) on %s access list.", i, mask);
                return;
            }
            i = z;

        } else {

           for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
               if (strcmp(*access, mask) == 0)
                    break;
           }
           if (i == ni->accesscount) {
               for (access = ni->access, i = 0; i < ni->accesscount;
                                                     ++access, ++i) {
               if (stricmp(*access, mask) == 0)
		    break;
               }
           }
           if (i == ni->accesscount) {
               notice(s_NickServ, source,
			"\2%s\2 not found on your access list.", mask);
               return;
           }
        }

        notice(s_NickServ, source,
             "\2%s\2 deleted from your access list.", *access);
        show_next_db(source, s_NickServ);
        if (readonly)
            notice(s_NickServ, source,
               "Warning: Services is in read-only mode; changes will not be saved.");

	free(*access);
	--ni->accesscount;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy(access+1, access, (ni->accesscount-i) * sizeof(char *));
	if (ni->accesscount)		/* if there are any entries left... */
	    ni->access = srealloc(ni->access, ni->accesscount * sizeof(char *));
	else {
	    free(ni->access);
	    ni->access = NULL;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_NickServ, source, "\2Access list:\2");
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (mask && !match_wild(mask, *access))
		continue;
            notice(s_NickServ, source, " %d)  %s", i+1, *access);
	}
	notice(s_NickServ, source, "\2End of List\2");

    } else if (stricmp(cmd, "WIPE") == 0) {
        for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	      free(*access);
	}
        ni->accesscount=0;
        free(ni->access);
        ni->access = NULL;
	notice(s_NickServ,source,"Your access list has been wiped.");

    } else {

	notice(s_NickServ, source,
                "Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    }
}

/*************************************************************************/

static void do_acc(const char *source)
{
 char *nick = strtok(NULL, " ");
 NickInfo *ni;

 if (!nick) {
     notice(s_NickServ, source, "Syntax: \2ACC\2 <nick>");
     notice(s_NickServ, source,
         "\2/msg %s HELP ACC\2 for more information.", s_NickServ);
          return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_VERBOTEN)) {
	notice(s_NickServ, source,
                "Nick \2%s\2: 5 [FORBIDDEN]", nick);
           return;

    } else if (!(ni = findnick(nick))) {
        notice(s_NickServ, source, "Nick \2%s\2: 0 [UnRegistered]", nick);
          return;

    } else if (!finduser(nick)) {
        notice(s_NickServ, source, "Nick \2%s\2: 1 [OffLine]", nick);
         return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_IDENTIFIED)) {
         notice(s_NickServ, source,
                 "Nick \2%s\2: 3 [Identified]", nick); 
           return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_RECOGNIZED)) {
        notice(s_NickServ, source, "\2%s\2: 4 Recognized [Access List]", nick);
           return;

    } else if ((ni = findnick(nick)) && !(ni->flags & NI_IDENTIFIED)) {
	notice(s_NickServ, source,
                "Nick \2%s\2: 2 [UnIdentified]", nick);
           return;


  } else {

     notice(s_NickServ, source, "Nick \2%s\2: Unable to Retreive Data", nick);
   }
}



/*************************************************************************/

static void do_info(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *tmp, timebuf[64];
    NickInfo *ni;
    time_t tbuff;
    User *u;

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2INFO\2 <nick>");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 isn't registered.", nick);

    } else if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, source,
		"Nick \2%s\2 may not be registered or used.", nick);

    } else {
	struct tm tm;
	char buf[512];
        u = finduser(nick);

        if (u && (ni->flags & NI_IDENTIFIED))
            notice(s_NickServ, source,
               "(\2Currently on IRC\2 [Identified]) - \"/whois %s\" for more information", nick);
        else if (u && (ni->flags & NI_RECOGNIZED))
            notice(s_NickServ, source,
               "(\2Currently on IRC\2 [Recognized]) - \"/whois %s\" for more information", nick);

        if (ni->last_realname)
            notice(s_NickServ, source,
	    	"\2%s\2 is \"%s\"\n", nick, ni->last_realname);
	notice(s_NickServ, source,
		"           Access: %s\n", get_os_access(ni));
	if (!(ni->flags & NI_NOHOST) && ni->last_usermask)
	    notice(s_NickServ, source,
		    "Last seen address: %s\n", ni->last_usermask);

	tm = *localtime(&ni->time_registered);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
        notice(s_NickServ, source,
                "  Time registered: %s", timebuf);

	tm = *localtime(&ni->last_seen);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
	notice(s_NickServ, source,
		"   Last seen time: %s", timebuf);
        tbuff = time(NULL);
        tm = *localtime(&tbuff);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
	notice(s_NickServ, source,
		"         Time Now: %s", timebuf);

	if (ni->url)
	    notice(s_NickServ, source,
		"              URL: %s\n", ni->url);
        if ((ni->email) && !(ni->flags & NI_HIDE_EMAIL))
	    notice(s_NickServ, source, "   E-mail address: %s\n", ni->email);
	else
	  if (is_oper(source))
	    notice(s_NickServ, source, "   E-mail address: %s\n", ni->email);


	*buf = 0;
	if (ni->flags & NI_KILLPROTECT)
	    strcat(buf, "Kill protection");
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
        if (ni->flags & NI_NEVEROP) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NEVER OP");
        }
        if (ni->flags & NI_NOOP) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO OP");
        }
        if (ni->flags & NI_NOMEMO) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO MEMO");
        }
        if (ni->flags & NI_NOHOST) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO HOST");
        }
        if (ni->flags & NI_HIDE_EMAIL) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Private");
        }

        if (ni->flags & NI_AUTH) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Auth");
        }

	if (!*buf)
	    strcpy(buf, "None");
	notice(s_NickServ, source, "          Options: %s", buf);

        if (ni->forward)
           notice(s_NickServ, source,
               " * Nick is forwarding memos to: %s", ni->forward);
        if (ni->flags & NI_HOP)
           notice(s_NickServ, source,
               " * Nick is Services Helpop");
        if (ni->flags & NI_SOP)
           notice(s_NickServ, source,
               " * Nick is Services Operator");

        if (is_oper(source)) {
            if (ni->flags & NI_MARK)
                notice(s_NickServ, source,
                   " * Nick marked by: %s", ni->mark);
            if (ni->flags & NI_HOLD)
                notice(s_NickServ, source,
                   " * Nick Held by: %s", ni->hold);
        } else {
           if (ni->flags & NI_MARK)
               notice(s_NickServ, source,
                  " * This nick has been marked");
           if (ni->flags & NI_HOLD)
               notice(s_NickServ, source,
                  " * This nick has been held");
       }        
               

    }
}

/*************************************************************************/

static void do_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    NickInfo *ni;
    int nnicks, i;
    char buf[BUFSIZE];
    int is_servadmin = is_services_admin(source);

    if (!pattern) {

        notice(s_NickServ, source, "Syntax: \2LIST\2 [pattern]");
	notice(s_NickServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_NickServ);

    } else {

	nnicks = 0;
	notice(s_NickServ, source, "List of entries matching \2%s\2:",
		pattern);
	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (ni->flags & NI_VERBOTEN)
		    continue;
		if (!is_servadmin && (ni->flags & NI_PRIVATE))
		    continue;
		snprintf(buf, sizeof(buf), "%-20s  %s", ni->nick, ni->last_usermask);
		if (stricmp(pattern, ni->nick) == 0 ||
					match_wild_nocase(pattern, buf)) {
		    if (++nnicks <= 50)
			notice(s_NickServ, source, "    %s", buf);
		}
	    }
	}
	notice(s_NickServ, source, "End of list - %d/%d matches shown.",
					nnicks>50 ? 50 : nnicks, nnicks);
    }

}

/*************************************************************************/

static void do_recover(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2RECOVER\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP RECOVER\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.");

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't recover yourself.");

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	collide(ni, 0);
	notice(s_NickServ, source, "User claiming your nick has been killed.");
	notice(s_NickServ, source,
		"\2/msg %s RELEASE\2 to get it back before the one-minute "
		"timeout.", s_NickServ);

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RECOVER: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RECOVER: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
           if (EXTRA_SNOOP) {
                slog("NS *R %s (%s!%@%s)",
                     nick, source, u->username, u->host);
                 do_break_log("NS_O", "NS *R %s (%s!%@%s)",
                     nick, source, u->username, u->host);
           }
	   bad_password(u, nick);
	}

    }
}

/*************************************************************************/

static void do_release(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2RELEASE\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP RELEASE\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!(ni->flags & NI_KILL_HELD)) {

	notice(s_NickServ, source, "Nick %s isn't being held.", nick);

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	release(ni, 0);
	notice(s_NickServ, source,
		"Services' hold on your nick has been released.");

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RELEASE: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RELEASE: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
            if (EXTRA_SNOOP) {
                slog("NS *Re %s (%s!%s@%s)",
                    nick, source, u->username, u->host);
                do_break_log("NS_O", "NS *Re %s (%s!%s@%s)",
                    nick, source, u->username, u->host);
            }
	    bad_password(u, nick);
	}

    }
}

/*************************************************************************/

static void do_ghost(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2GHOST\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP GHOST\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.", nick);

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't ghost yourself!");

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	char buf[NICKMAX+32];

	snprintf(buf, sizeof(buf), "Ghost command used by %s", source);
        kill_user(SERVER_NAME, nick, buf);
	notice(s_NickServ, source, "Ghost with your nick has been killed.");

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: GHOST: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: GHOST: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
            if (EXTRA_SNOOP) {
                slog("NS *K %s (%s!%s@%s)",
                     nick, source, u->username, u->host);
                do_break_log("NS_O", "NS *K %s (%s!%s@%s)",
                     nick, source, u->username, u->host);
            }
	    bad_password(u, nick);
	}

    }
}

/*************************************************************************/

static void do_sendpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

#ifdef USE_ENCRYPTION
    notice(s_NickServ, source,
		"GETPASS command unavailable when encryption is in use.");
#else
    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2SENDPASS\2 <nickname>");
    } else if (!u) {
	notice(s_NickServ, source, "Couldn't get your user info!");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (ni->flags & NI_MARK) {
        slog("NS *S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        do_break_log("NS_G", "NS *S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
            notice(s_NickServ,source,"Nick \2%s\2 is marked!",nick);
        log("%s: %s!%s@%s attempted SENDPASS on %s (Fail)"
                ,s_NickServ, source, u->username, u->host, nick);

    } else if (!ni->email) {
          notice(s_NickServ, source,
               "User %s doesn't have an E-Mail specified", ni->nick);

    } else {
	log("%s: %s!%s@%s used SENDPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
        slog("NS S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        slog("NS S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        wallops(s_NickServ, "%s used SENDPASS on %s", source, nick);
        notice(s_NickServ, source,
              "Password sent to %s", ni->email);
        sendpass_ns(source, ni);
    }
#endif
}


/*************************************************************************/

static void do_getpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

#ifdef USE_ENCRYPTION
    notice(s_NickServ, source,
		"GETPASS command unavailable when encryption is in use.");
#else
    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2GETPASS\2 <nickname>");
    } else if (!u) {
	notice(s_NickServ, source, "Couldn't get your user info!");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick %s isn't registered.", nick);
    } else if (is_services_admin(nick) && !is_services_coder(source)
          && !stricmp(source, nick) == 0) {
        notice(s_NickServ, source,
           "You may not getpass a services administrator");

    } else if (ni->flags & NI_MARK) {
        slog("NS *G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        do_break_log("NS_G", "NS *G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
            notice(s_NickServ,source,"Nick \2%s\2 is marked!",nick);
        log("%s: %s!%s@%s attempted GETPASS on %s (Fail)"
                ,s_NickServ, source, u->username, u->host, nick);
    } else {
	log("%s: %s!%s@%s used GETPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
        slog("NS G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        slog("NS G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        wallops(s_NickServ, "%s used GETPASS on %s", source, nick);
	notice(s_NickServ, source, "Password for %s is \2%s\2.",
		nick, ni->pass);
    }
#endif
}

/*************************************************************************/

static void do_forbid(const char *source)
{
    NickInfo *ni;
    User *u = finduser(source);
    char *nick = strtok(NULL, " ");

    if (!is_services_oper(source)) {
	notice(s_NickServ,source,"Permission Denied.");
	return;
    }

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2FORBID\2 [-]<nickname>");
        notice(s_NickServ,source,
             "\2/msg %s OHELP FORBID\2 for more information",s_NickServ);
	return;
    }
    if (*nick == '-') {
       *nick++;
       do_unforbid(source, nick);
       return;
    }
    if (readonly) {
	notice(s_NickServ, source,
	    "Warning: Services is in read-only mode; changes will not be saved.");
    }
    if (is_services_oper(nick)) {
           notice(s_NickServ, source, "Error: Nick is a services admin");
           wallops(s_NickServ, "%s tried to forbid \2%s\2", source, nick);
           return;
    }
    if (ni = findnick(nick)) {
        if (ni->forbid) {
           notice(s_NickServ, source, "Nick %s already forbidden", nick);
           return;
        }
	delnick(ni);
    }
    if (ni = makenick(nick)) {
	ni->flags |= NI_VERBOTEN;
        if (ni->forbid)
             free(ni->forbid);
        ni->forbid = sstrdup(u->nick);
        slog("NS +F %s (%s!%s@%s)",
             nick,u->nick,u->username,u->host);
        do_break_log("NS_F", "NS +F %s (%s!%s@%s)",
             nick,u->nick,u->username,u->host);
	log("%s: %s set FORBID for nick %s", s_NickServ, source, nick);
        wallops(s_NickServ, "%s set \2%s\2 as FORBIDDEN ", source, nick);
	notice(s_NickServ, source, "Nick \2%s\2 is now FORBIDden.", nick);
    } else {
	log("%s: Valid FORBID for %s by %s failed", s_NickServ,
		nick, source);
	notice(s_NickServ, source, "Couldn't FORBID nick \2%s\2!", nick);
    }
}

static void do_unforbid(const char *source, const char *nick)
{
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2FORBID\2 [-]<nick>");
        notice(s_NickServ, source,
            "\2/MSG %s OHELP FORBID\2 for more information", s_NickServ);
	return;
    }
    if (readonly) {
	notice(s_NickServ, source,
	    "Warning: Services is in read-only mode; changes will not be saved.");
    }

   if (ni = findnick(nick)) {
       if (!(ni->flags & NI_VERBOTEN)) {
           slog("NS -F* %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
           do_break_log("NS_F", "NS -F* %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);

           log("%s: %s tried to set unforbidden flag for non forbidden nick %s",
                   s_NickServ, source, nick);
           if (ni->forbid)
               free(ni->forbid);
           ni->forbid = NULL;

           notice(s_NickServ, source, "%s is not set as FORBIDDEN", nick);
           return;
        }

        slog("NS -F %s (%s!%s@%s) [Dropped]",
             nick,u->nick,u->username,u->host);
        do_break_log("NS_F", "NS -F %s (%s!%s@%s) [Dropped]",
             nick,u->nick,u->username,u->host);
        log("%s: %s set FORBID for nick %s", s_NickServ, source, nick);
        delnick(ni);
        wallops(s_NickServ, "%s removed \2%s\2 from the FORBIDDEN list [Automatically Dropped]", source, nick);
        notice(s_NickServ, source, "Nick %s is now UNFORBIDDEN.", nick);
    } else {
        log("%s: Valid UNFORBID for %s by %s failed", s_NickServ,
		nick, source);
        notice(s_NickServ, source, "Couldn't UNFORBID nick \2%s\2!", nick);
    }
}

static char *get_os_access(NickInfo *ni)
{
    char *access, buf[128];
    int i=0, end=0;
   

    for (i = 0; i < MAX_SERVROOTS; i++) {
        if (services_roots[i]) {
           if (stricmp(services_roots[i], ni->nick) == 0) {
              end++;
              if (stricmp(services_roots[i], SERVICES_MASTER) == 0)
                   snprintf(buf, sizeof(buf), "Services Root [Master], IRC Operator");
	      else if (stricmp(services_roots[i], SERVICES_MASTER2) == 0)
                   snprintf(buf, sizeof(buf), "Services Root [Co Master], IRC Operator");
              else
                   snprintf(buf, sizeof(buf), "Services Root, IRC Operator");
           }
        }
    }

    for (i = 0; i < MAX_SERVADMINS; i++) {
        if (services_admins[i]) {
           if (stricmp(services_admins[i], ni->nick) == 0) {
                   if (!end) {
                      snprintf(buf, sizeof(buf), "Services Admin, IRC Operator");
                      end++;
                   }
           }
        }
    }
   if (ni->flags & NI_SOP)
      if (!end) {
      	snprintf(buf, sizeof(buf), "Services Operator, IRC Operator");
        end++;
      }

   if (ni->flags & NI_HOP)
      if (!end) {
        if (is_oper(ni->nick))
	        snprintf(buf, sizeof(buf), "IRC Operator");
        else
        	snprintf(buf, sizeof(buf), "Help Operator");
        end++;
      }

/*
    for (i = 0; i < MAX_SERVOPERS; i++) {
        if (services_opers[i]) {
           if (stricmp(services_opers[i], ni->nick) == 0) {
                   if (!end) {
                      snprintf(buf, sizeof(buf), "Services Operator, IRC Operator");
                      end++;
                   }
           }
        }
    }

    for (i = 0; i < MAX_HELPOPS; i++) {
        if (services_helpers[i]) {
           if (stricmp(services_helpers[i], ni->nick) == 0) {
                   if (!end) {
                      if (is_oper(ni->nick))
                          snprintf(buf, sizeof(buf), "IRC Operator");
                      else
                          snprintf(buf, sizeof(buf), "Help Operator");
                      end++;
                   }
           }
        }
    }
*/
    if (is_oper(ni->nick) && !end) {
         snprintf(buf, sizeof(buf), "IRC Operator");
         end++;
    }

    if (!end)
       snprintf(buf, sizeof(buf), "User");

    access = buf;
    return access;
}

static void do_flist(const char *source)
{
    char *cmd = strtok(NULL, " ");
    NickInfo *ni;
    int i;

    if (!is_services_oper(source)) {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);
        return;
    }

    if (!cmd) {
        notice(s_NickServ,source,"Syntax: \2FLIST\2 [HELD|MARKED|FORBIDDEN]");
        notice(s_NickServ,source,
            "\2/msg %s OHELP FLIST\2 for more information",s_NickServ);

    } else if (stricmp(cmd, "HELD") == 0) {
        notice(s_NickServ,source,"\2HELD\2 nickname listing:");
        for  (i = 0; i < 256; ++i) {
          for (ni = nicklists[i]; ni; ni = ni->next) {
              if (ni->hold) {
                notice(s_NickServ,source,"%-20s  %s",
                   ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->hold);
              }
          }
        }
        notice(s_NickServ,source,"\2End of List\2");
    } else if (stricmp(cmd, "MARKED") == 0) {
        notice(s_NickServ,source,"\2MARKED\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->mark) {
                  notice(s_NickServ,source,"%-20s  %s",
                        ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->mark);
                }
            }
        }
        notice(s_NickServ,source,"\2End of List\2");
    } else if (stricmp(cmd, "FORBIDDEN") == 0) {
        notice(s_NickServ,source,"\2FORBIDDEN\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->forbid) {
                  notice(s_NickServ,source,"%-20s  %s",
                        ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->forbid);
                }
            }
        }
        notice(s_NickServ,source,"\2End of List\2");
    } else if (stricmp(cmd, "HOP") == 0) {
        notice(s_NickServ,source,"\2HOP\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_HOP) {
                  notice(s_NickServ,source,"%-20s",
                        ni->nick);
                }
            }
        }
        notice(s_NickServ,source,"\2End of List\2");
    } else if (stricmp(cmd, "SOP") == 0) {
        notice(s_NickServ,source,"\2SOP\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_SOP) {
                  notice(s_NickServ,source,"%-20s",
                        ni->nick);
                }
            }
        }
        notice(s_NickServ,source,"\2End of List\2");
    }
}

/*************************************************************************/

static void do_unhold(const char *source, const char *nick)
{
     User *u = finduser(source);
     NickInfo *ni;

     if (!(ni = findnick(nick))) {
          notice(s_NickServ, source, "Nickname %s is not registered.",nick);
     } else if (ni->flags & NI_HOLD) {
          ni->flags &= ~NI_HOLD;
         if (ni->hold)
             free(ni->hold);
          ni->hold = NULL;
          slog("NS -H %s (%s!%s@%s)",
              nick,u->nick,u->username,u->host);
          do_break_log("NS_F", "NS -H %s (%s!%s@%s)",
               nick,u->nick,u->username,u->host);
          wallops(s_NickServ,
              "%s unset the HELD flag for \2%s\2",u->nick,nick);
     } else {
          notice(s_NickServ,source,"Nickname \2%s\2 is not held.",nick);
     }
}

/*************************************************************************/

static void do_unmark(const char *source, const char *nick)
{
     User *u = finduser(source);
     NickInfo *ni;

     if (!(ni = findnick(nick))) {
          notice(s_NickServ, source, "Nickname %s is not registered.",nick);
     } else if (ni->flags & NI_MARK) {
          ni->flags &= ~NI_MARK;
         if (ni->mark)
             free(ni->mark);
          ni->mark = NULL;
          slog("NS -M %s (%s!%s@%s)",
              nick,u->nick,u->username,u->host);
          do_break_log("NS_F", "NS -M %s (%s!%s@%s)",
               nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
              "%s unset the MARK flag for \2%s\2",u->nick,nick);
     } else {
          notice(s_NickServ,source,"Nickname \2%s\2 is not marked.",nick);
     }

}


/*************************************************************************/

static void do_setpass(const char *source)
{
    char *nick = strtok (NULL, " ");
    char *newpass = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick)
        notice (s_NickServ, source, "Syntax: \2SETPASS Nick Password\2");
    else if (!u)
        notice (s_NickServ, source, "Where are you?");

    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, "nick %s is not registered. ", nick);

    else if ((ni->flags & NI_MARK) && (!is_services_admin(source)))
        notice (s_NickServ, source, "Can not SETPASS on marked nick. ", nick);
    else
    {
       strscpy (ni->pass, newpass, PASSMAX);
       show_next_db(source, s_NickServ);
       notice(s_NickServ, source, "Password for nick %s changed to \2%s\2.", nick, newpass);
    }
}

/**************************************************************************/

static void do_set_hidemail(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param,"ON") == 0) {

         if (!ni->email) {
            notice(s_NickServ, ni->nick,
                "You have no email to HIDE");
            return;
         }

         ni->flags |= NI_HIDE_EMAIL;
         notice(s_NickServ, ni->nick,
                  "Your email address will no longer be shown");
         show_next_db(source, s_NickServ);

    } else if (stricmp(param,"OFF") == 0) {

       if (!ni->email) {
           notice(s_NickServ, ni->nick,
               "You have no email address set to show!");
           return;
       }

       ni->flags &= ~NI_HIDE_EMAIL;
       notice(s_NickServ, source, "Your EMAIL is no longer hide");
       show_next_db(source, s_NickServ);

     } else  {
           notice(s_NickServ, source, "Syntax: SET HIDEMAIL ON/OFF");
     }
}
/************************************************************************/
static void do_set_auth(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "You need to authorized before add to any channel access");
       ni->flags |= NI_AUTH;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "Disable authorized.");
       ni->flags &= ~NI_AUTH;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET AUTH\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET AUTH\2 for more information.",
              s_NickServ);

  }
}
/***************************************************************************/
static void do_set_privmsg(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "You may recieve notice in private message");
       ni->flags |= NI_PRIVMSG;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "You may recieve services message in notice.");
       ni->flags &= ~NI_PRIVMSG;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET PRIVMSG\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET PRIVMSG\2 for more information.",
              s_NickServ);

  }
}
/**************************************************************************/
static void do_set_malay(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "Change help language to Malay");
       ni->flags |= NI_MALAY;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "Change help language to English.");
       ni->flags &= ~NI_MALAY;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET MALAY\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET MALAY\2 for more information.",
              s_NickServ);

  }
}
/******************************************************************/
static void do_setflags(const char *source)
{
    char *nick = strtok (NULL, " ");
    char *nflag = strtok (NULL, " ");
    char *noption = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !nflag || !noption)
        notice (s_NickServ, source, "Syntax: \2SETFLAG Nick flag ON/OFF\2");
    else if (!u)
        notice (s_NickServ, source, "Where are you?");

    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, "nick %s is not registered. ", nick);

    else if ((ni->flags & NI_MARK) && (!is_services_admin(source)))
        notice (s_NickServ, source, "Can not SETFLAG on marked nick. ", nick);

    else if (!(stricmp(noption, "ON")==0) && !(stricmp(noption, "OFF")==0))
        notice (s_NickServ, source, "Option must be ON or OFF");

    else
    {
        if (stricmp(nflag, "AUTH")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_AUTH;
		else 
		       ni->flags &= ~NI_AUTH;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "HOP")==0) {
		if(stricmp(noption, "ON")==0) {
		       ni->flags &= ~NI_SOP;
		       ni->flags |= NI_HOP;
		} else 
		       ni->flags &= ~NI_HOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);
	
	} else if (stricmp(nflag, "SOP")==0) {
		if(stricmp(noption, "ON")==0) {
		       ni->flags &= ~NI_HOP;
		       ni->flags |= NI_SOP;
		} else 
		       ni->flags &= ~NI_SOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "MALAY")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_MALAY;
		else 
		       ni->flags &= ~NI_MALAY;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "PRIVMSG")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_PRIVMSG;
		else 
		       ni->flags &= ~NI_PRIVMSG;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "EMAILMEMOS")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_EMAILMEMOS;
		else 
		       ni->flags &= ~NI_EMAILMEMOS;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "NOOP")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_NOOP;
		else 
		       ni->flags &= ~NI_NOOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);
	}	
    }
}
/***********************************************************************/
#endif	/* !SKELETON */


